/*
   This code is from "Algorithms in Java, Third Edition,"
   by Robert Sedgewick, Addison-Wesley, 2003.
*/
#include "InsertionSort.h"

#define exch(A, B) { int t = A; A = B; B = t; }



// Programs 7.1, 7.2, page 317, 319
void quicksort1(int a[], int l, int r)
{
   if (r <= l)
   {
      return;
   }
   else
   {
      int v = a[r];
      int i = l-1, j = r;
      for (;;)
      {
         while (a[++i] < v) ;
         while (v < a[--j]) if (j == l) break;
         if (i >= j) break;
         exch(a[i], a[j]);
      }
      exch(a[i], a[r]);

      quicksort1(a, l,   i-1);
      quicksort1(a, i+1, r);
   }
}




// Program 7.4, Improved quicksort, page 332,
// (with median-of-three partitioning)
void quicksort2Helper(int*, int, int);

void quicksort2(int a[], int l, int r)
{
   quicksort2Helper(a, l, r);
      insertionSort(a, l, r);
}


#define M 10

void quicksort2Helper(int a[], int l, int r)
{
   if (r-l <= M)
   {
      return;
   }
   else
   {
      int v, i, j;
      // compute median-of-three pivot
      exch(a[(l+r)/2], a[r-1]);
      if (a[l]   < a[r-1]) exch(a[l],   a[r-1]);
      if (a[l]   < a[r])   exch(a[l],   a[r]);
      if (a[r-1] < a[r])   exch(a[r-1], a[r]);
      // now do the partition
      v = a[r];
      i = l, j = r-1;
      for (;;)
      {
         while (a[++i] < v) ;
         while (v < a[--j]) if (j == l) break;
         if (i >= j) break;
         exch(a[i], a[j]);
      }
      exch(a[i], a[r]);

      quicksort2Helper(a, l,   i-1);
      quicksort2Helper(a, i+1, r);
   }
}




// Program 7.5, Quicksort with three-way partitioning, page 338
void quicksort3Helper(int*, int, int);

void quicksort3(int a[], int l, int r)
{
   quicksort3Helper(a, l, r);
      insertionSort(a, l, r);
}


void quicksort3Helper(int a[], int l, int r)
{
   if (r-l <= M)
   {
      return;
   }
   else
   {
      int v = a[r];
      int i = l-1, j = r, p = l-1, q = r, k;
      for (;;)
      {
         while (a[++i] < v) ;
         while (v < a[--j]) if (j == l) break;
         if (i >= j) break;
         exch(a[i], a[j]);
         if (v == a[i]) { p++; exch(a[p], a[i]); }
         if (v == a[j]) { q--; exch(a[q], a[j]); }
      }
      exch(a[i], a[r]);
      // Now a[l..p] = v,  a[p+1..i-1] < v, a[i] = v, a[i+1..q-1] > v , a[q..r] = v
      j = i-1; i = i+1;
      for (k = l  ; k <= p; k++, j--) exch(a[k], a[j]);
      for (k = r-1; k >= q; k--, i++) exch(a[k], a[i]);
      // Now a[l..j] < v = a[j+1..i-1] < a[i..r]

      quicksort3Helper(a, l, j);
      quicksort3Helper(a, i, r);
   }
}




// This last example is from "Algorithms, Fourth Edition,"
// by Robert Sedgewick, Addison-Wesley, 2011, page 299.
// This is also Quicksort with three-way partitioning, but
// this version should be a bit slower than the previous one.
void quicksort4Helper(int*, int, int);

void quicksort4(int a[], int l, int r)
{
   quicksort4Helper(a, l, r);
      insertionSort(a, l, r);
}


void quicksort4Helper(int a[], int l, int r)
{
   if (r-l <= M)
   {
      return;
   }
   else
   {
      int lt = l, i = l+1, gt = r;
      int v = a[l];
      while (i <= gt)
      {
         if      (a[i] < v) { exch(a[lt], a[i]); lt++; i++; }
         else if (a[i] > v) { exch(a[i], a[gt]); gt--; }
         else               { i++; }
      } // Now a[l..lt-1] < v = a[lt..gt] < a[gt+1..r]

      quicksort4Helper(a, l,    lt-1);
      quicksort4Helper(a, gt+1, r);
   }
}
